import { AfterViewInit, ChangeDetectorRef, Component, Input, NgZone, OnInit, ViewChild } from '@angular/core';
import { NgxMoveableComponent, OnDragEnd, OnResize, OnResizeEnd } from 'ngx-moveable';
import { OnDrag } from 'ngx-moveable';
import { ConfigService } from '../_services/config.service';
import { Rect } from '../_models/generated/Rect';
import { ShortcutService } from '../_services/shortcut.service';
import { CarouselComponent } from '../carousel/carousel.component';
import { PlanService, UrlError } from '../_services/plan.service';
import { Plan, PlanElement, PlanMetadata } from '../_models/plan';
import { WorldAreaService } from '../_services/world-area.service';
import { WorldArea } from '../_models/world-area';
import { Subscription, from, timer } from 'rxjs';
import { open } from '@tauri-apps/api/dialog';
import { OverlayService, StateEvent } from '../_services/overlay.service';
import { appWindow } from '@tauri-apps/api/window';
import { EventsService } from '../_services/events.service';
import { Event } from '@tauri-apps/api/event';
import { MatDialog } from '@angular/material/dialog';
import { ResumeDialog } from './resume-dialog.component';
import { TimeTrackerService } from '../_services/time-tracker.service';

enum Resume {
  Discard,
  Next,
  Instant
}

@Component({
  selector: 'plan-display',
  templateUrl: './plan-display.component.html',
  styleUrls: ['./plan-display.component.scss']
})
export class PlanDisplayComponent implements AfterViewInit, OnInit {
  @Input() backgroundColor?: String;
  draggable: boolean = true;
  rect?: Rect;
  bounds: any = { "left": 0, "top": 0, "right": 0, "bottom": 0, "position": "css" };
  @ViewChild("moveable") moveable?: NgxMoveableComponent;

  slideIndex: number = 0;
  zoneSlides?: CarouselComponent<PlanElement>;
  currentSlides?: CarouselComponent<PlanElement>;
  worldAreaMap?: Map<String, WorldArea>;
  settingsOpen: boolean = false;
  init: boolean = false;
  hasAttachedOnce: boolean = false;

  overlayStateChangeHandle?: Subscription;
  bindsAreSetup: boolean = false;
  nextBind?: Subscription;
  prevBind?: Subscription;

  currentPlan?: Plan;
  previousPlans: PlanMetadata[] = [];
  checkingPlanUpdate: boolean = false;

  recentUpdateAttempts: Map<string, Date | UrlError> = new Map<string, Date | UrlError>();

  constructor(private events: EventsService, public configService: ConfigService, private cdr: ChangeDetectorRef, private shortcut: ShortcutService, public planService: PlanService, public worldAreaService: WorldAreaService, public overlayService: OverlayService, private zone: NgZone, public dialog: MatDialog, public timeTrackerService: TimeTrackerService) {
    window.addEventListener("resize", () => {
      this.zone.run(() => {
        this.windowInitHandler()
      })
    });

    this.planService.getStoredPlans().subscribe(plans => {
      console.log("got new stored plans");
      this.previousPlans = plans;
    })

    this.planService.getCurrentPlan().subscribe(plan => {
      this.currentPlan = plan;

      this.timeTrackerService.onNewRun(plan);
      //Close settings anytime we get a new current plan.
      this.settingsOpen = false;

      setTimeout(() => this.setIndex(plan.current), 0);
    })

    this.registerOnZoneEnter();
  }

  get disablePlans(): boolean {
    return this.checkingPlanUpdate;
  }

  registerOnZoneEnter() {
    appWindow.listen("entered", (entered) => {
      if (this.currentPlan) {
        const current = this.currentPlan.current;
        const length = this.currentPlan.plan.length;
        if (current + 1 < length) {
          if (entered.payload === this.currentPlan.plan[current + 1].area_key) {
            this.zone.run(() => this.next());
          }
        }
      }
    });
  }

  windowInitHandler() {
    if (window.innerWidth > 0) {
      this.ngAfterViewInit();
    }
  }

  ngOnInit() {
    this.worldAreaService.getWorldAreas().subscribe(a => this.worldAreaMap = a);
    this.overlayStateChangeHandle = this.events.listen<StateEvent>("OverlayStateChange").subscribe(this.onOverlayStateChange.bind(this));
  }

  onOverlayStateChange(event: Event<StateEvent>) {
    if (event.payload.Hidden) {
      this.destroyBinds();
    } else {
      this.setupBinds();
    }
  }

  destroyBinds() {
    if (this.bindsAreSetup) {
      this.nextBind?.unsubscribe();
      this.prevBind?.unsubscribe();
      this.bindsAreSetup = false;
    }
  }

  abs(v: number) {
    return Math.abs(v);
  }

  transform() {
    return `translate(${this.rect!.x}px, ${this.rect!.y}px)`;
  }

  width() {
    return `${this.rect!.width}px`;
  }

  height() {
    return `${this.rect!.height}px`;
  }

  hasWaypoint(key?: string): boolean {
    if (!key) {
      key = this.currentPlan!.plan[this.currentPlan!.current].area_key;
    }
    const world_area = this.worldAreaMap?.get(key);
    return world_area!.has_waypoint;
  }

  hasTrial(key?: string): boolean {
    if (!key) {
      key = this.currentPlan!.plan[this.currentPlan!.current].area_key;
    }

    return this.worldAreaService.hasTrial(key);
  }

  ngAfterViewInit(): void {
    if (window.innerWidth > 0) {
      const cfgRect = this.configService.config.initialPlanWindowPosition;

      this.rect = {
        x: cfgRect.x * window.innerWidth,
        y: cfgRect.y * window.innerHeight,
        width: cfgRect.width * window.innerWidth,
        height: cfgRect.height * window.innerHeight,
      }
      this.moveable?.updateRect();

      setTimeout(() => this.cdr.detectChanges(), 0);
      this.init = true;
    }
  }

  onDrag(e: OnDrag) {
    this.rect!.x = e.translate[0];
    this.rect!.y = e.translate[1];
  }

  onDragEnd(e: OnDragEnd) {
    this.saveRect();
  }

  onResize(e: OnResize) {
    this.rect!.width = e.width;
    this.rect!.height = e.height;
    this.onDrag(e.drag);
  }

  onResizeEnd(e: OnResizeEnd) {
    this.saveRect();
  }

  saveRect() {
    const toCfgRect = this.rect!;
    this.configService.config.initialPlanWindowPosition = {
      x: toCfgRect.x / window.innerWidth,
      y: toCfgRect.y / window.innerHeight,
      width: toCfgRect.width / window.innerWidth,
      height: toCfgRect.height / window.innerHeight,
    }
  }

  registerZoneSlides(carousel: CarouselComponent<PlanElement>) {
    this.zoneSlides = carousel;
    this.zoneSlides.setIndex(this.slideIndex);
  }

  setupBinds() {
    if (this.currentSlides && !this.bindsAreSetup) {
      
      this.nextBind = this.shortcut.register(this.configService.config.prev).subscribe((_shortcut) => {
        this.prev();
        if (this.configService.config.enableStopwatch) {
          this.timeTrackerService.onForcePrev(this.currentPlan!.plan[this.currentPlan!.current].area_key);
        }
      });

      this.prevBind = this.shortcut.register(this.configService.config.next).subscribe((_shortcut) => {
        this.next();
        if (this.configService.config.enableStopwatch) {
          this.timeTrackerService.onForceNext(this.currentPlan!.plan[this.currentPlan!.current].area_key);
        }
      });
      
      this.bindsAreSetup = true;
    }
  }

  registerCurrentSlides(carousel: CarouselComponent<PlanElement>) {
    this.currentSlides = carousel;
    this.currentSlides.setIndex(this.slideIndex);

    this.setupBinds();
  }

  next() {
    if (this.overlayService.visible) {
      this.currentPlan!.next();
      this.currentSlides?.next();
      this.zoneSlides?.next();
    }
  }

  prev() {
    if (this.overlayService.visible) {
      this.currentPlan!.prev();
      this.currentSlides?.prev();
      this.zoneSlides?.prev();

    }
  }

  setIndex(index: number) {
    this.slideIndex = index;
    if (this.currentSlides) {
      this.currentSlides.setIndex(index);
    }
    if (this.zoneSlides) {
      this.zoneSlides.setIndex(index);
    }
  }

  loadPrevious(path: string) {
    this.planService.loadPlanFromPath(path, false).subscribe(plan => {
      this.planService.setCurrentPlan(plan);
    });
  }

  settingsClick(event: any) {
    this.settingsOpen = !this.settingsOpen;
    event.stopPropagation();
  }

  openDialog() {
    from(open({
      multiple: false,
      filters: [
        {
          name: "JSON (.json)",
          extensions: ['json']
        }
      ]
    })).subscribe(file => {
      if (file) {
        this.planService.loadPlanFromPath(file as string).subscribe(plan => {
          if (plan) {
            this.planService.setCurrentPlan(plan);
          }
        });
      }
    });
  }

  loadBasePlan() {
    this.planService.getBasePlan().subscribe(plan => {
      this.currentPlan = new Plan(plan);
      if (this.zoneSlides) {
        this.zoneSlides.setIndex(0);
      }
      if (this.currentSlides) {
        this.currentSlides.setIndex(0);
      }
    })
  }

  onScroll(event: WheelEvent) {
    if (event.deltaY < 0) {
      this.prev();
      this.timeTrackerService.onForcePrev(this.currentPlan!.plan[this.currentPlan!.current].area_key);
    } else {
      this.next();
      this.timeTrackerService.onForceNext(this.currentPlan!.plan[this.currentPlan!.current].area_key);
    }
  }

  specialClasses() {
    const waypoint = this.hasWaypoint() ? 'active' : '';
    const trial = this.hasTrial() ? 'trial-active' : '';
    return `${waypoint} ${trial}`;
  }

  clampedOffset(): number {
    return Math.min(this.configService.config.numVisible - 1, this.configService.config.offset);
  }

  zonesStyle() {
    return {
      'min-height': `${this.configService.config.numVisible * 18}px`,
      'max-height': `${this.configService.config.numVisible * 40}px`
    }
  }

  loadFromUrl() {
    this.planService.loadFromUrl().subscribe(plan => {
      if (plan) {
        this.planService.savePlanAtStore(plan.name!, plan).subscribe((path) => {
          if (path) {
            plan.setPath(path);
          }
        });
        this.planService.setCurrentPlan(plan);
      }
    });
  }

  checkForPlanUpdate(plan: PlanMetadata) {
    this.checkingPlanUpdate = true;

    this.planService.checkForPlanUpdate(plan).subscribe({
      next: updatedPlan => {
        console.log("check update: ", updatedPlan);
        this.planService.loadPlanFromPath(plan.stored_path!, false).subscribe(oldPlan => {
          console.log("oldPlan", oldPlan);
          const newCurrent = commonUUIDToNewCurrent(oldPlan, updatedPlan);
          updatedPlan.current = newCurrent;
          //TODO: Interface to determine if we want to save the plan or not (allow not doing it in case user regrets... Also needs position error fix).
          this.planService.savePlanAtStore(oldPlan.name!, updatedPlan, true).subscribe();
        })
      },
      complete: () => {
        this.checkingPlanUpdate = false
        this.recentUpdateAttempts.set(plan.update_url!, new Date());
      },
      error: (err) => {
        this.recentUpdateAttempts.set(plan.update_url!, err);
        if (err instanceof UrlError) {
          alert("Error retrieving plan update (" + err.status + "): " + err.message);
        } else if (err instanceof Error) {
          alert("Unexpected error: " + err.message);
        }

        this.checkingPlanUpdate = false;
      }
    });
  }

  isErr(plan: PlanMetadata) {
    return this.recentUpdateAttempts.get(plan.update_url!) instanceof UrlError;
  }

  hasUpdate(plan: PlanMetadata): any {
    return this.recentUpdateAttempts.has(plan.update_url!);
  }

  recentPlanTitle(plan: PlanMetadata) {
    if (!this.hasUpdate(plan)) return "Check for update";
    if (this.hasUpdate(plan) && !this.isErr(plan)) return "Up to date";
    if (this.hasUpdate(plan) && this.isErr(plan)) return "Error updating";
    return "";
  }

  shouldDisplayTimer(): boolean {
    if (!this.configService.config.enableStopwatch) return false;

    return this.timeTrackerService.isActive;
  }

  stopStopwatch() {
    this.timeTrackerService.stop();
  }

  startStopwatch() {
    this.timeTrackerService.startLoaded();
  }
}

function commonUUIDToNewCurrent(oldPlan: Plan, updatedPlan: Plan): number {
  let bestOldCurrent = oldPlan.current;
  let newIndex = 0;
  //Attempt current plan element's UUID first, otherwise step backwards until we find a match
  while (bestOldCurrent >= 0) {
    const tempNewIndex = updatedPlan.plan.findIndex(e => e.uuid === oldPlan.plan[bestOldCurrent].uuid);
    if (tempNewIndex !== -1) {
      newIndex = tempNewIndex;
      break;
    }
    bestOldCurrent--;
  }

  return newIndex;
}
