import { Component, OnInit } from '@angular/core';
import { CommonModule } from '@angular/common';
import { RunHistory, RunHistoryMetadata, TimeTrackerService } from '../_services/time-tracker.service';
import { MatTableModule } from '@angular/material/table';
import { MatSlideToggleModule } from '@angular/material/slide-toggle';
import { FormsModule } from '@angular/forms';
import { ScrollingModule } from '@angular/cdk/scrolling';
import { MatButtonModule } from '@angular/material/button';
import { WorldAreaService } from '../_services/world-area.service';
import { WorldArea } from '../_models/world-area';

export interface RunStat {
  zoneName: string;

  entryTime: string;
  estimatedExit: string;
  estimatedTimeSpent: string;
}

export interface AggregateRunStat {
  zoneName: string;

  aggregateFirstEntry: string;
  aggregateLastExit: string;
  aggregateTimeSpent: string;
  aggregateNumEntries: string;
}

interface UnformattedAggregateRunStat {
  zoneId: string;

  aggregateFirstEntry: number;
  aggregateLastExit: number;
  aggregateTimeSpent: number;
  aggregateNumEntries: number;
}

type RunStatType = RunStat | AggregateRunStat;

@Component({
  selector: 'app-run-stats',
  standalone: true,
  imports: [CommonModule, MatTableModule, MatSlideToggleModule, FormsModule, ScrollingModule, MatButtonModule],
  templateUrl: './run-stats.component.html',
  styleUrls: ['./run-stats.component.scss']
})
export class RunStatsComponent implements OnInit {
  aggregated?: AggregateRunStat[];
  /// practically which zone can't have a last exit time as last exit is not determinable for the last entry
  aggregateNAId?: string;

  direct?: RunStat[];

  aggregate: boolean = true;
  cache?: Map<string, RunHistoryMetadata>;
  worldAreaMap?: Map<string, WorldArea>;

  constructor(private timeTrackerService: TimeTrackerService, private worldAreaService: WorldAreaService) {
    this.worldAreaService.getFullWorldAreas().subscribe((data) => {
      this.worldAreaMap = data;
    })
  }

  ngOnInit(): void {
    this.timeTrackerService.storedHistories.subscribe((data) => {
      this.cache = data;
    })
  }

  get dataSource(): RunStatType[] {
    return (this.aggregate ? this.aggregated : this.direct) ?? [];
  }

  get hasInitializedData() {
    return this.aggregated || this.direct;
  }

  get shouldShowTable() {
    return this.hasInitializedData;
  }

  loadMain(uuid: string) {
    this.timeTrackerService.loadHistory(uuid).subscribe((data => {
      if (data) {
        this.onLoad(data);
      }
    }));
  }

  dateFormat(value: number) {
    return new Date(value).toLocaleString();
  }

  onLoad(data: RunHistory) {
    this.direct = this.calcDirect(data);
    this.aggregated = this.calcAggregated(data);
  }

  calcAggregated(data: RunHistory): AggregateRunStat[] {
    const aggregation = new Map<string, UnformattedAggregateRunStat>();

    this.aggregateNAId = data.entries[data.entries.length - 1].zone;

    data.entries.forEach((entry, index) => {
      const hasExit = !(data.entries.length - 1 === index);

      let aggregate: UnformattedAggregateRunStat = {
        zoneId: entry.zone,
        aggregateFirstEntry: entry.current_elapsed_millis,
        aggregateLastExit: hasExit ? data.entries[index + 1].current_elapsed_millis : 0,
        aggregateTimeSpent: hasExit ? (data.entries[index + 1].current_elapsed_millis - data.entries[index].current_elapsed_millis) : 0,
        aggregateNumEntries: 1,
      }

      const existing = aggregation.get(entry.zone);
      if (existing) {
        existing.aggregateLastExit = aggregate.aggregateLastExit;
        existing.aggregateTimeSpent += aggregate.aggregateTimeSpent;
        existing.aggregateNumEntries++;
      }

      aggregation.set(entry.zone, existing ?? aggregate);
    });

    return Array.from(aggregation.values()).map((entry) => {
      let aggregateTimeSpent;
      if (this.aggregateNAId === entry.zoneId) {
        aggregateTimeSpent = this.timeTrackerService.hmsTimestamp(entry.aggregateTimeSpent) + " + N/A"
      } else {
        aggregateTimeSpent = this.timeTrackerService.hmsTimestamp(entry.aggregateTimeSpent);
      }

      return {
        zoneName: this.resolveZone(entry.zoneId),
        aggregateFirstEntry: this.timeTrackerService.hmsTimestamp(entry.aggregateFirstEntry),
        aggregateLastExit: this.aggregateNAId === entry.zoneId ? "N/A" : this.timeTrackerService.hmsTimestamp(entry.aggregateLastExit),
        aggregateTimeSpent: aggregateTimeSpent,
        aggregateNumEntries: entry.aggregateNumEntries.toString(),
      }
    });

  }

  calcDirect(data: RunHistory): RunStat[] {
    return data.entries.map((entry, index) => {
      const hasExit = !(data.entries.length - 1 === index);
      return {
        zoneName: this.resolveZone(entry.zone),
        entryTime: this.timeTrackerService.hmsTimestamp(entry.current_elapsed_millis),
        estimatedExit: hasExit ? this.timeTrackerService.hmsTimestamp(data.entries[index + 1].current_elapsed_millis) : "N/A",
        estimatedTimeSpent: hasExit ? this.timeTrackerService.hmsTimestamp(data.entries[index + 1].current_elapsed_millis - data.entries[index].current_elapsed_millis) : "N/A",
      }
    })
  }

  hms(time: number) {
    return this.timeTrackerService.hmsTimestamp(time);
  }

  private resolveZone(zoneId: string) {
    const area = this.worldAreaMap?.get(zoneId);
    if (!area) {
      return "Unknown zone: " + zoneId;
    }

    // Act might not be very reasonable but it's the best we have it..
    return area.name + " (A" + area.act + ")"
  }

  reset() {
    this.aggregated = undefined;
    this.direct = undefined;
  }

  get displayedColumns() {
    if (this.aggregate) {
      return [
        "zoneName",
        "aggregateFirstEntry",
        "aggregateLastExit",
        "aggregateTimeSpent",
        "aggregateNumEntries",
      ]
    } else {
      return [
        "zoneName",
        "entryTime",
        "estimatedExit",
        "estimatedTimeSpent",
      ]
    }
  }
}
